--[[
********************************
Licence: GPL
(C)2012 rudi, Version: 0.5

NMEA GPS LIBRARY
********************************
]]

nmeaLib = {}

nmeaLib.version = "0.5"

local GpsPower --boolean

-- begin: VALID CAMERA TABLE
local function getCamValues()

    -- List of valid cameras
    -- !!! each entry for camera and firmware
    local cam_tab = {}

    cam_tab[#cam_tab+1] = {platform="sx260hs", platsub="100b", aGpsStart=0xff04b6e0, aGpsStop=0xff04b6f0, pNmeaBuffer=0x65ac+0x8}
    cam_tab[#cam_tab+1] = {platform="sx260hs", platsub="100c", aGpsStart=0xff04b6e0, aGpsStop=0xff04b6f0, pNmeaBuffer=0x65ac+0x8}

    local res
    local i, bi = 1, get_buildinfo()
    while (i <= #cam_tab) do
        if (bi.platform == cam_tab[i].platform) and (bi.platsub == cam_tab[i].platsub) then
            res = cam_tab[i]
            break
        end
        i = i + 1
    end
    return res
end

local camValues = getCamValues()
-- end: VALID CAMERA TABLE

-- begin: NMEA READ AND VERIFY BUFFER DATA
local function verifyNmeaChecksumAndTrim(msg)
    local res, msgres = -1, ""
    if (#msg > 1 + 5) then
        local sum = tonumber(string.sub(msg, -4,-3), 16)
        if sum then
            msgres = string.sub(msg, 4, -6)
            local msgsum = string.byte(msg, 2)
            for i = 3, #msg - 5 do
                msgsum = bitxor(msgsum, string.byte(msg, i))
            end
            if sum == msgsum then
                res = 1
            else
                res = 0
                msgres = ""
            end
        end
    end
    return msgres, res
end

--[[
    read block from nmea buffer max. 8 lines with max. 82 bytes
    $GPGGA, $GPGSA, $GPRMC, $GPVTG and max. 4x $GPGSV
]]
local function readBuffer()
    local pbuffer, buffer, lines = peek(camValues.pNmeaBuffer), "", 8
    local lineOffset, i, cr, b = 0, 0, 0, 0x00
    while (i < 82) and (cr < lines) do
        b = peek(pbuffer + lineOffset + i, 1)
        if (i == 0) and (b ~= 0x24) then break end  --need $ as first char
        buffer = buffer..string.char(b)
        i = i + 1
        if b == 0x0A then   -- stop on LF
            cr = cr + 1
            lineOffset = lineOffset + i
            i = 0
        end
    end
    return buffer
end

--[[
    get each nmea line from buffer and store in tData
]]
local function readData()
    local res, tres, msg = 0
    local buffer, valid = readBuffer()
    if string.sub(buffer, 1, 3) == "$GP" then   --NMEA:GPS
        for msg in string.gmatch(buffer, "$GP%u%u%u[%u%d%.,%-]-*%x%x\r\n") do
            msg, valid = verifyNmeaChecksumAndTrim(msg)
            if valid == 1 then
                if not tres then tres = {} end
                tres[#tres+1] = msg
                if string.sub(msg, 1, 3) == "VTG" then
                    GpsPower = true
                    res = 1
                    break
                end
            else
                break
            end
        end
    else
        msg = string.match(buffer, "$PMTK%d%d%d[%d%,]-*%x%x\r\n")
        if msg then
            msg, valid = verifyNmeaChecksumAndTrim(msg)
            if valid == 1 then
                if string.sub(msg, 1, 5) == "TK001" then
                    --any power acknowledge
                    GpsPower = false
                elseif string.sub(msg, 1, 5) == "TK010" then
                    --"power on" acknowledge
                    GpsPower = true
                end
                tres = {}
                res = 1
            end
        end
    end
    return tres, res
end

--[[
    read new nmea data
    return:
        nil = error
         {} = nmea data
]]
function nmeaLib.getData()
    local i, res, tres = 0, 0
    -- read again on error,
    -- e.g. while camera gps task work with buffer
    while (res == 0) and (i < 10) do
        if i > 0 then sleep(50) end
        i = i + 1
        tres, res = readData()
    end
    if res == 0 then tres = nil end
    return tres
end
-- end: NMEA READ AND VERIFY BUFFER DATA

-- begin: CONVERT NMEA DATA TO GPS TABLE
local function getTabFromNmea(msg, vals)
    local tres = {}
    for part in string.gmatch(msg..",", "([%w%.%-]-),") do    --split on ","
        tres[#tres+1] = part
    end
    if (vals > 0) and (#tres ~= vals) then tres = {} end
    return tres
end

local function intStrToNumber(str)
    return tonumber("0"..str)
end

local function getTabFromNmeaGsv(Msg, tgsv, List)
    local t = getTabFromNmea(Msg, 0)
    if not tgsv then tgsv = {} end
    if t and (#t%4 == 0) and (#t/4 > 1) then
        for i = 1, #t/4 - 1 do
            if List then
                for j = 1, #List do
                    if List[j]==intStrToNumber(t[4*i+1]) then
                        tgsv[#tgsv+1] = {intStrToNumber(t[4*i+1]),
                                         intStrToNumber(t[4*i+2]),
                                         intStrToNumber(t[4*i+3]),
                                         intStrToNumber(t[4*i+4])}
                    end
                end
            elseif intStrToNumber(t[4*i+4]) > 0 then
                tgsv[#tgsv+1] = {intStrToNumber(t[4*i+1]),
                                 intStrToNumber(t[4*i+2]),
                                 intStrToNumber(t[4*i+3]),
                                 intStrToNumber(t[4*i+4])}
            end;
        end
    end
    if #tgsv == 0 then tgsv = nil end
    return tgsv
end

local function r64u_1(value)
    local a, b = string.match(value, "(%d+)%.(%d+)")
    if a and b then
        return {tonumber(a..b), 10^#b}
    end
end

local function r64u_3time(value)
    local h, m, s, ms = string.match(value, "(%d%d)(%d%d)(%d%d)%.(%d%d%d)")
    if h and m and s and ms then
        return {tonumber(h), 1, tonumber(m), 1, tonumber(s..ms), 1000}
    end
end

local function r64u_3geo(value)
    local d, m, s = string.match(value, "(%d-%d%d)(%d%d)%.(%d%d%d%d)")
    if d and m and s then
        return {tonumber(d), 1, tonumber(m), 1, tonumber(s*6), 1000}
    end
end

local function i8u_4(value)
    local v1, v2, v3, v4 = string.match(value, "(%d)(%d)(%d)(%d)")
    if v1 and v2 and v3 and v4 then
        return {tonumber(v1), tonumber(v2), tonumber(v3), tonumber(v4)}
    end
end

local function i8u_1(value)
    local v = string.match(value, "(%d+)")
    if v and (tonumber(v) < 256) then return tonumber(v) end
end

local function str_1(value)
    return string.match(value, "(%w)")
end

local function str_11date(value)
    local dd, mm, yy = string.match(value, "(%d%d)(%d%d)(%d%d)")
    if dd and mm and yy then
        if tonumber(yy) < 80 then yy = "20"..yy else yy = "19"..yy end
        return string.format("%s:%s:%s", yy, mm, dd)
    end
end

local function addExifEntry(tab, key, value, func)
    if tab[key] or (#value == 0) then return end
    if func == "r64u_3t" then tab[key] = r64u_3time(value)
    elseif func == "r64u_3" then tab[key] = r64u_3geo(value)
    elseif func == "r64u_1" then tab[key] = r64u_1(value)
    elseif func == "i8u_4" then tab[key] = i8u_4(value)
    elseif func == "i8u_1" then tab[key] = i8u_1(value)
    elseif func == "str_1" then tab[key] = str_1(value)
    elseif func == "str_11d" then tab[key] = str_11date(value)
    end
end

local function nmeaToGpstab(tMsgList)
    local tres, tMSG = {}, {}
    local cmd
    for i = 1, #tMsgList do
        cmd = string.sub(tMsgList[i], 1, 3)
        if cmd == "GGA" then
            tMSG = getTabFromNmea(tMsgList[i], 15)
            if tMSG and (tMSG[7] ~= "0") then
                addExifEntry(tres, "TimeStamp", tMSG[2], "r64u_3t")
                addExifEntry(tres, "Latitude", tMSG[3], "r64u_3")
                addExifEntry(tres, "LatitudeRef", tMSG[4], "str_1")
                addExifEntry(tres, "Longitude", tMSG[5], "r64u_3")
                addExifEntry(tres, "LongitudeRef", tMSG[6], "str_1")
                addExifEntry(tres, "FixQuality", tMSG[7], "str_1")
                addExifEntry(tres, "SatToTrack", tMSG[8], "i8u_1")
                addExifEntry(tres, "Altitude", tMSG[10], "r64u_1")
                local ar = "0"
                if string.sub(tMSG[10], 1, 1) == "-" then ar = "1" end
                addExifEntry(tres, "AltitudeRef", ar, "i8u_1")
            end
        elseif cmd == "GSA" then
            tMSG = getTabFromNmea(tMsgList[i], 18)
            if tMSG then
                addExifEntry(tres, "MeasureMode", tMSG[3], "str_1")
                if tres.SatToTrack and (tres.SatToTrack>0) then
                    local t = {}
                    for j = 1, tres.SatToTrack do
                        addExifEntry(t, j, tMSG[3+j], "i8u_1")
                    end
                    tres["SatTrackList"] = t
                end
            end
        elseif cmd == "GSV" then
            tres["SatInfo"] = getTabFromNmeaGsv(tMsgList[i], tres["SatInfo"], tres["SatTrackList"])
        elseif cmd == "RMC" then
            tMSG = getTabFromNmea(tMsgList[i], 13)
            if tMSG  then
                addExifEntry(tres, "Status", tMSG[3], "str_1")
                addExifEntry(tres, "VersionID", "2300", "i8u_4")
                if (tMSG[13] ~= "N") then
                    addExifEntry(tres, "TimeStamp", tMSG[2], "r64u_3t")
                    addExifEntry(tres, "Latitude", tMSG[4], "r64u_3")
                    addExifEntry(tres, "LatitudeRef", tMSG[5], "str_1")
                    addExifEntry(tres, "Longitude", tMSG[6], "r64u_3")
                    addExifEntry(tres, "LongitudeRef", tMSG[7], "str_1")
                    addExifEntry(tres, "DateStamp", tMSG[10], "str_11d")
                end
            end
        elseif cmd == "VTG" then
            tMSG = getTabFromNmea(tMsgList[i], 10)
            if tMSG then
                addExifEntry(tres, "VersionID", "2300", "i8u_4")
                if tMSG[10] ~= "N" then
                    addExifEntry(tres, "Track", tMSG[2], "r64u_1")
                    addExifEntry(tres, "TrackRef", tMSG[3], "str_1")
                    addExifEntry(tres, "Speed", tMSG[8], "r64u_1")
                    addExifEntry(tres, "SpeedRef", tMSG[9], "str_1")
                end
            end
        end
    end
    for k in pairs(tres) do     --any elements in tres?
        tres["ProcessingMethod"] = "GPS"
        tres["MapDatum"] = "WGS-84"
        break
    end
    return tres
end

--[[
    convert nmea table to exif equal gps table
    use a given nmea table or new read nmea table
]]
function nmeaLib.getGpsTable(value)
    local tres = value
    if type(value) ~= "table" then
        tres = nmeaLib.getData()
    end
    if tres then
        return nmeaToGpstab(tres)
    end
end
-- end: CONVERT NMEA DATA TO GPS TABLE

-- begin: FORMAT FUNCTIONS FOR GPS TABLE
function nmeaLib.getLatitudeStr(t)
    local res = [[-  -- --' --"]]
    if t and t.LatitudeRef and t.Latitude and (#t.Latitude == 6) then
        res = string.format([[%s  %2d %02d' %02d"]], t.LatitudeRef,
                    t.Latitude[1]/t.Latitude[2],
                    t.Latitude[3]/t.Latitude[4],
                    t.Latitude[5]/t.Latitude[6])
    end
    return res
end

function nmeaLib.getLongitudeStr(t)
    local res = [[- --- --' --"]]
    if t and t.LongitudeRef and t.Longitude and (#t.Longitude == 6) then
        res = string.format([[%s %3d %02d' %02d"]], t.LongitudeRef,
                    t.Longitude[1]/t.Longitude[2],
                    t.Longitude[3]/t.Longitude[4],
                    t.Longitude[5]/t.Longitude[6])
    end
    return res
end

function nmeaLib.getAltitudeStr(t)
    local res = [[-----.-m]]
    if t and t.Altitude and (#t.Altitude == 2) and t.AltitudeRef then
        local alt = 10 * t.Altitude[1]/t.Altitude[2]
        if t.AltitudeRef == 1 then
            res = "-"..string.format([[%4d.%dm]], alt/10, alt%10)
        else
            res = string.format([[%5d.%dm]], alt/10, alt%10)
        end
    end
    return res
end

function nmeaLib.getSpeedStr(t)
    local res = [[---.--km/h]]
    if t and t.Speed and (#t.Speed == 2) then
        local speed = 100*t.Speed[1]/t.Speed[2]
        res = string.format([[%3d.%02dkm/h]], speed/100, speed%100)
    end
    return res
end

function nmeaLib.getTrackStr(t)
    local res = [[---]]
    if t and t.Track and (#t.Track == 2) then
        res = string.format([[%3d]], t.Track[1]/t.Track[2])
    end
    return res
end

function nmeaLib.getDateStampStr(t)
    local res = [[--/--/----]]
    if t and t.DateStamp then
        local yyyy, mm, dd = string.match(t.DateStamp, "(%d%d%d%d):(%d%d):(%d%d)")
        if yyyy and mm and dd then
            res = string.format([[%s/%s/%s]], dd, mm, yyyy)
        end
    end
    return res
end

function nmeaLib.getTimeStampStr(t)
    local res = [[--:--:--]]
    if t and t.TimeStamp and (#t.TimeStamp == 6) then
        res = string.format([[%2d:%02d:%02d]], t.TimeStamp[1]/t.TimeStamp[2],
                                               t.TimeStamp[3]/t.TimeStamp[4],
                                               t.TimeStamp[5]/t.TimeStamp[6])
    end
    return res
end
-- end: FORMAT FUNCTIONS FOR GPS TABLE

-- begin: NMEA GPS POWER
--  for all: gps power [off/on] = [false/true]
--  GpsPower is set by function readData()

local function waitForGpsPowerStatusEnd(value)
    local i = 15    --max. 3s
    while (i > 0) and (GpsPower == value) do
        sleep(200)
        i = i - 1
        readData()
    end
end

--[[
    init gps power (switch gps power to off on unknown status)
]]
local function initGpsPower()
    if not GpsPower then
        call_func_ptr(camValues.aGpsStart)
        waitForGpsPowerStatusEnd(false)
        call_func_ptr(camValues.aGpsStop)
        waitForGpsPowerStatusEnd(true)
    end
    return GpsPower
end

--[[
    return current gps power status
]]
function nmeaLib.getGpsPower()
    readData()
    return GpsPower or initGpsPower()
end

--[[
    set gps power
    return: current gps power status
]]
function nmeaLib.setGpsPower(value)
    if (GpsPower ~= value) then
        -- changing is required
        if value == true then
            call_func_ptr(camValues.aGpsStart)
            waitForGpsPowerStatusEnd(false)
        else
            call_func_ptr(camValues.aGpsStop)
            waitForGpsPowerStatusEnd(true)
        end
    end
    return GpsPower
end
-- end: NMEA GPS POWER

-- begin: LIBRARY RESULT
--[[
    library need OPT_LUA_CALL_NATIVE compiler option and
    an entry with valid addresses in camera list
]]
if camValues then
    if call_func_ptr then
        --ready, return library as table
        return nmeaLib
    else
        --missing function call_func_ptr()
        return 1
    end
else
    --no entry in camera list
    return -1
end
-- end: LIBRARY RESULT
